package bcfg

  1. Overview
  2. Docs

bcfg

bcfg

API

Library bcfg

  • Bcfg bcfg is a simple configuration format built on two types: the directive and the list of directives. The format is meant to be small enough to be read by a human, and regular enough to be processed by a program without any ceremony.
  • Bcfg_json A small, dependency-free JSON view of a bcfg configuration, useful to pipe a query result into tools such as jq.
  • Bcfg_lexer
  • Bcfg_out
  • Bcfg_parser
  • Bcfg_parser_messages
  • Bcfg_stream Fully streaming, SAX-like API for bcfg.
  • Bcfg_txtloc
  • Bcfg_type

Library bcfg.cli

Library bcfg.iso

Library bcfg.lint

Library bcfg.q

Library bcfg.query

Library bcfg.validate